\ The compiler
: ALLOT ( n -- )                            \ Allocate n bytes to the code dictionary.
  CP +! ;

: , ( w -- )                                \ Compile an integer into the code dictionary.
  HERE ALIGNED DUP CELL+ CP ! ! ;

: CALL, ( ca -- )                           \ Assemble a call instruction to ca.
  HERE CELL+ CELL+ - 
  CELL/ FFFFFF AND [ =CALL ] LITERAL OR , ; 

: [COMPILE] ( -- ; <string> )               \ Compile the next immediate word into code dictionary.
  [hidden] ' [TARGET] CALL, ; IMMEDIATE

: COMPILE ( -- )                            \ Compile the next address in colon list to code dictionary.
  R> DUP @ CELLS OVER + HERE - 
  CELL/ FFFFFF AND [ =CALL ] LITERAL OR , 
  CELL+ >R ; COMPILE-ONLY

ONLY META ALSO TARGET ALSO hidden DEFINITIONS
: LITERAL 
  COMPILE doLIT , ; IMMEDIATE
ONLY META ALSO TARGET ALSO DEFINITIONS

: $," ( -- )                                \ Compile a literal string up to next " .
  -4 ALLOT
  [ CHAR " ] LITERAL WORD COUNT + ALIGNED CP ! ;

: RECURSE ( -- )                            \ Make the current word available for compilation.
  LAST @ NAME> , ; IMMEDIATE

\ Name compiler
: ?UNIQUE ( a -- a )                        \ Display a warning message if the word already exists.
  DUP NAME? 
    IF ."  reDef " OVER .$
    THEN 
  DROP ;

: $,n ( na -- )                             \ Build a new dictionary name using the string at na.
  DUP C@
  IF ?UNIQUE
    ( na) DUP NAME> CP !
    ( na) DUP LAST ! \ for OVERT
    ( na) CELL-
    ( la) CURRENT @ @
    ( la voc) SWAP !
    EXIT
  THEN ABORT" name" ;

\ FORTH compiler
: $COMPILE ( a -- )                         \ Compile next word to code dictionary as a token or literal.
  NAME? ?DUP
  IF @ [ =IMMEDIATE ] LITERAL AND
    IF EXECUTE ELSE CALL, THEN EXIT
  THEN
  'NUMBER @EXECUTE
  IF [hidden] LITERAL [TARGET] EXIT
  THEN THROW ;

ONLY META ALSO TARGET ALSO hidden DEFINITIONS
: ; 
  [TARGET] [ E8B28000 ] LITERAL ,  
  [hidden] [ [TARGET] OVERT ; COMPILE-ONLY IMMEDIATE

[hidden]
: ]
  [TARGET] [ ' $COMPILE ] LITERAL 'EVAL ! ;

[hidden]
: :
  [TARGET] TOKEN $,n [ E9224000 ] LITERAL , [hidden] ] [TARGET] ;

\ Defining words

ONLY META ALSO TARGET ALSO DEFINITIONS
: USER ( u -- ; <string> )                  \ Compile a new user variable.
  TOKEN $,n [ E9224000 ] LITERAL , OVERT COMPILE doUSER , ;

: CREATE ( -- ; <string> )                  \ Compile a new array entry without allocating code space.
  TOKEN $,n [ E9224000 ] LITERAL , OVERT COMPILE doVAR ;

: VARIABLE ( -- ; <string> )                \ Compile a new variable initialized to 0.
  CREATE 0 , ;

: CONSTANT ( -- ; <string> )                  \ Compile a new array entry without allocating code space.
  TOKEN $,n [ E9224000 ] LITERAL , OVERT COMPILE doCON , ;

\ Structures
: FOR ( -- a )                              \ Start a FOR-NEXT loop structure in a colon definition.
  COMPILE >R HERE ; IMMEDIATE

: BEGIN ( -- a )                            \ Start an infinite or indefinite loop structure.
  HERE ; IMMEDIATE

: NEXT ( a -- )                             \ Terminate a FOR-NEXT loop structure.
  COMPILE next , ; IMMEDIATE

: UNTIL ( a -- )                            \ Terminate a BEGIN-UNTIL indefinite loop structure.
  COMPILE ?branch , ; IMMEDIATE

: AGAIN ( a -- )                            \ Terminate a BEGIN-AGAIN infinite loop structure.
  COMPILE  branch , ; IMMEDIATE

: IF ( -- A )                               \ Begin a conditional branch structure.
  COMPILE ?branch HERE 0 , ; IMMEDIATE

: AHEAD ( -- A )                            \ Compile a forward branch instruction.
  COMPILE branch HERE 0 , ; IMMEDIATE

: REPEAT ( A a -- )                         \ Terminate a BEGIN-WHILE-REPEAT indefinite loop.
  AGAIN HERE SWAP ! ; IMMEDIATE

: THEN ( A -- )                             \ Terminate a conditional branch structure.
  HERE SWAP ! ; IMMEDIATE

: AFT ( a -- a A )                          \ Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.
  DROP AHEAD BEGIN SWAP ; IMMEDIATE

: ELSE ( A -- A )                           \ Start the false clause in an IF-ELSE-THEN structure.
  AHEAD SWAP THEN ; IMMEDIATE

: WHILE ( a -- A a )                        \ Conditional branch out of a BEGIN-WHILE-REPEAT loop.
  IF SWAP ; IMMEDIATE

: EXIT ( -- )                                  \ Terminate a colon definition.
  [ E8B28000 ] LITERAL , ; IMMEDIATE

: ABORT"      ( -- ; <string> )             \ Conditional abort with an error message.
  COMPILE abort"| $," ; IMMEDIATE

: $" ( -- ; <string> )                      \ Compile an inline string literal.
  COMPILE $"| $," ; IMMEDIATE

: ." ( -- ; <string> )                      \ Compile an inline string literal to be typed out at run time.
  COMPILE ."| $," ; IMMEDIATE

: IMMEDIATE ( -- )                          \ Make the last compiled word an immediate word.
  [ =IMMEDIATE ] LITERAL LAST @ C@ OR LAST @ C! ;


